import unittest

from restler.entity import Entity


class Fields_for_Simple_Object(unittest.TestCase):

    def setUp(self):
        self.entity = Entity()
        self.default_set = set([('id', 'id'), ('id_str', 'id_str')])

    def should_be_default_set_when_input_is_None(self):
        fields = self.entity._parse_fields_for_simple_object(None)
        self.assertEqual(fields, self.default_set)

    def should_be_default_set_when_input_is_glob(self):
        fields = self.entity._parse_fields_for_simple_object(['*'])
        self.assertEqual(fields, self.default_set)

    def should_include_extra_fields_when_specified(self):
        fields = self.entity._parse_fields_for_simple_object(
            ['*', '+_public_names'])
        expected_fields = self.default_set.union(
            set([('_public_names', '_public_names')]))
        self.assertEqual(fields, expected_fields)

    def should_exclude_specified_fields(self):
        fields = self.entity._parse_fields_for_simple_object(['*', '-id_str'])
        expected_fields = self.default_set - set([('id_str', 'id_str')])
        self.assertEqual(fields, expected_fields)

    def should_work_with_legacy_dict_format(self):
        fields = self.entity._parse_fields_for_simple_object({
            '*': '*', 
            'a': 'a',
            '+b': 'B',
            'b': 'B',
            '-id_str': '-id_str'
        })
        expected_fields = self.default_set.union(set([('a', 'a'), ('b', 'B')]))
        expected_fields -= set([('id_str', 'id_str')])
        self.assertEqual(fields, expected_fields)


class TestController(unittest.TestCase):
    pass



import logging

from paste.deploy.converters import asbool, aslist

from pylons import request, response, url
from pylons import tmpl_context as c
from pylons.controllers import WSGIController
from pylons.controllers.util import abort, redirect
from pylons.decorators import jsonify
from pylons.templating import render_mako as render

from sqlalchemy.orm import class_mapper

import mako.exceptions

try:
    import json
except ImportError:
    import simplejson as json


log = logging.getLogger(__name__)

TemplateNotFoundExceptions = (mako.exceptions.TopLevelLookupException,)


class NoDefaultValue(object):

    def __new__(self, *args, **kwargs):
        raise NotImplementedError('This class may not be instantiated.')


class Controller(WSGIController):

    entity = None
    """Entity class assocated with this controller."""

    base_filter_params = dict(
        where_clause=NoDefaultValue,  # Any valid SQL where clause fragment
        distinct=False,
        offset=None,
        start=None,
        limit=None,
        order_by=None,
    )

    filter_params = {}
    """Request param names with defaults, for filtering collections."""

    filters = []
    """SQLAlchemy filters--anything that can be an arg to `query().filter`.

    These filters are *not* request-specific--they are applied on *every*
    request.

    """

    default_format = 'json'

    def __call__(self, environ, start_response):
        try:
            return super(Controller, self).__call__(environ, start_response)
        finally:
            log.debug('Clearing database session...')
            self.db_session.remove()

    def __before__(self, *args, **kwargs):
        self.db_session.get_bind(class_mapper(self.entity))
        route_info = request.environ['pylons.routes_dict']
        self.controller = route_info['controller']
        self.action = route_info['action']
        self.member_name = self.entity.member_name
        self.member_title = self.entity.member_title
        self.collection_name = self.entity.collection_name
        self.collection_title = self.entity.collection_title
        self.format = kwargs.get('format',
            request.params.get('format', self.default_format))
        self._init_properties()
        log.debug('Action: %s' % self.action)

    def _get_db_session(self):
        """Database session factory.

        The default implementation here assumes that subclasses have a public
        :meth:`get_db_session` method. The session returned from that method
        is cached per :class:`Controller` instance.

        """
        try:
            self._db_session
        except AttributeError:
            self._db_session = self.get_db_session()
        return self._db_session
    def _set_db_session(self, db_session):
        self._db_session = db_session
    db_session = property(_get_db_session, _set_db_session)

    def get_db_session(self):
        raise NotImplementedError

    @property
    def collection_path(self):
        """Path to collection, including application path prefix."""
        try:
            self._collection_path
        except AttributeError:
            self._collection_path = '/'.join((
                request.path.rsplit(self.collection_name, 1)[0].rstrip('/'),
                self.collection_name))
        return self._collection_path

    def get_member_path(self, member):
        """Path to member, including application path prefix."""
        id = getattr(member, 'id_str', None)
        return '{0}/{1}'.format(self.collection_path, id)

    def index(self):
        self.set_collection()
        return self._render()

    def show(self, id):
        self.set_member(id)
        return self._render()

    def new(self):
        self.set_member()
        return self._render()

    def edit(self, id):
        self.set_member(id)
        return self._render()

    def create(self):
        self.set_member()
        self._update_member_with_params()
        self.db_session.add(self.member)
        self.db_session.flush()
        self.db_session.commit()
        self._redirect_to_member()

    def update(self, id):
        self.set_member(id)
        self._update_member_with_params()
        self.db_session.flush()
        self.db_session.commit()
        self._redirect_to_member()

    def delete(self, id):
        self.set_member(id)
        self.db_session.delete(self.member)
        self.db_session.flush()
        self.db_session.commit()
        self._redirect_to_collection()

    def set_member(self, id=None):
        if id is None:
            member = self.entity()
        else:
            member = self.get_entity_or_404(id)
        self.member = member

    def set_collection(self, q=None, extra_filters=None, filter_params=None):
        q = q if q is not None else self.db_session.query(self.entity)

        # Apply "global" (i.e., every request) filters
        filters = (self.filters or []) + (extra_filters or [])
        for f in filters:
            q = q.filter(f)

        # Apply per-request filters
        filters = self._set_filters_from_params(self.base_filter_params)
        filters.update(self._set_filters_from_params(self.filter_params))
        if filter_params is not None:
            filters.update(self._set_filters_from_params(filter_params))

        distinct = asbool(filters.pop('distinct', False))
        offset = filters.pop('offset', filters.pop('start', None))
        limit = filters.pop('limit', None)
        order_by = filters.pop('order_by', None)

        where_clause = filters.pop('where_clause', NoDefaultValue)
        if where_clause is not NoDefaultValue:
            q = q.filter(where_clause)

        for k, v in filters.items():
            v = self.convert_param(k, v)
            filter_method = getattr(self.entity, 'filter_by_%s' % k, None)
            if filter_method is not None:
                q = filter_method(q, v)
            else:
                q = q.filter_by(**{k: v})

        if distinct:
            q = q.distinct()
        if order_by is not None:
            q = q.order_by(*aslist(order_by, ','))
        if offset is not None:
            q = q.offset(int(offset))
        if limit is not None:
            q = q.limit(int(limit))

        self.collection = q.all() or abort(404)

    def _set_filters_from_params(self, filter_params):
        filters = {}
        params = request.params
        for name in filter_params:
            if name in params:
                # Get value from request params. If the value is not blank,
                # that value is used; otherwise, we use the default value.
                val = params.get(name)
                if not val:
                    val = filter_params[name]
            else:
                # Param wasn't passed in request params, so use the default
                # value.
                val = filter_params[name]
            if val is not NoDefaultValue:
                if callable(val):
                    val = val()
                filters[name] = val
        return filters

    def get_entity_or_404(self, id):
        id = self.entity.str_to_id(id)
        entity = self.db_session.query(self.entity).get(id) or abort(404)
        return entity

    def _update_member_with_params(self):
        params = request.params
        for name in params:
            val = self.convert_param(name, params[name])
            setattr(self.member, name, val)

    def convert_param(self, name, val):
        """Convert param value (string) to Python value."""
        return self.entity.convert_param(name, val)

    def _do_redirect(self, url_args, redirect_args=None):
        _url_args = dict(
            controller=self.controller,
            format=self.format,
        )
        _url_args.update(url_args)
        redirect_url = url(**_url_args)
        if request.is_xhr:
            # X-Restler-Client-Request-URL is the URL used by the client to
            # make requests to the Web service that is utilizing Restler.
            # This is only relevant when using AJAX due to same-origin
            # restrictions.
            base_url = request.headers.get('X-Restler-Client-Request-URL', None)
            if base_url is not None:
                redirect_url = base_url.rstrip('/') + redirect_url
        _redirect_args = dict(code=303)
        _redirect_args.update(redirect_args or {})
        redirect(redirect_url, **_redirect_args)

    def _redirect_to_member(self, member=None, relay_params=None, params=None):
        """Redirect to a specific ``member``, defaulting to `self.member`.

        ``relay_params`` List of params to relay from original request. If
        a param isn't in the original request, it's ignored and not relayed
        to the redirect.

        ``params`` Dict of additional params; will override ``relay_params``.

        """
        member = self.member if member is None else member
        url_args = {}
        if relay_params:
            for key in relay_params:
                if key in request.params:
                    url_args[key] = request.params[key]
        if params:
            for key in params:
                url_args[key] = params[key]
        url_args.update(action='show', id=member.id)
        self._do_redirect(url_args)

    def _redirect_to_collection(self):
        self._do_redirect(dict(action='index'))

    def _render(self, *args, **kwargs):
        format = kwargs.get('format', self.format)
        log.debug('Output format: %s' % format)
        kwargs['format'] = format
        render = getattr(self, '_render_%s' % format, self._render_template)
        log.debug('Render method: %s *%s **%s' % (render.__name__, args, kwargs))
        response.status = kwargs.pop('code', 200)
        return render(*args, **kwargs)

    def _render_template(
        self, controller=None, action=None, format=None, namespace=None):
        """By default, render template /{controller}/{action}.{format}.

        This is the default rendering method, used when a specific
        ``render_<format>`` method doesn't exist. Templates can be any
        text-based template--HTML, text, XML, etc.

        Typically, args are set in the environment (on ``c``), except
        ``namespace``. Args passed explicitly will override args set in the
        env. If ``namespace`` is given, the template rendered will be
        /namespace/{controller}/{action}.{format}.

        """
        template = '/%%s/%s.%s' % (action or self.action, format or self.format)
        try:
            template_name = template % (controller or self.controller)
            return render(template_name)
        except TemplateNotFoundExceptions:
            if namespace is not None:
                template = '%s/%s' % (namespace, template)
            template_name = template % 'default'
            return render(template_name)
        finally:
            log.debug('(_render) template: %s' % template_name)

    def _render_json(self, block=None, **kwargs):
        """Render a JSON response from simplified ``member``s."""
        obj = self._get_json_object(wrap=self.wrap, block=block)
        return self._render_object_as_json(obj)

    @jsonify
    def _render_object_as_json(self, obj):
        """Render an object in JSON format with correct content type.

        ``obj`` must be JSONifiable by the (simple)json module.

        The final output of this method is returned by the Pylons ``jsonify``
        decorator, which creates a proper JSON response with the correct
        content type.

        """
        return obj

    def _get_json_object(self, wrap=True, block=None):
        """Get JSON object for current request.

        ``wrap``
             If set, the output will be wrapped in a dict.

        ``block``
            Can be passed to modify or wrap the object before JSONifying it.
            In this case the wrapping discussed above under ``obj`` won't
            happen.

        """
        if self.collection is not None:
            log.debug('Rendering collection')
            items = self.collection
        elif self.member is not None:
            log.debug('Rendering member')
            items = [self.member]
        else:
            log.debug('Neither collection nor member was set.')
            items = None
            obj = None
            result_count = 0

        if items is not None:
            obj = self.entity.to_simple_collection(items, self.fields)
            for member, simple_member in zip(items, obj):
                simple_member['__path__'] = self.get_member_path(member)
            result_count = len(obj)

        # Wrap ``obj`` (usually)
        if wrap:
            obj = dict(
                response=dict(
                    results=obj,
                    result_count=result_count,
                    request=dict(
                        method=request.method,
                        full_url=request.url,  # URL with query string
                        host_url=request.host_url,  # URL of host (no path or query)
                        app_prefix=request.script_name,  # Path to app
                        path=request.path,  # Path *including* app prefix
                        collection_path=self.collection_path,  # *Includes* app prefix
                        params=(request.params.items() or None),
                        query_string=(request.query_string or None),
                    ),
                )
            )
        # Further modify ``obj`` if ``block`` given
        if block is not None:
            obj = block(obj)
        return obj

    @property
    def fields(self):
        """Return list of fields to include in response.

        The `fields` request parameter should be a JSON `list`. This property
        merely decodes the parameter from JSON into a Python object and
        returns it. See :class:`restler.entity.Entity` for documentation on
        the expected form and contents of the list.

        Example (URL-encoded)::

            /path?fields=["*","%2Bmy_attr"]

            Decoded from JSON: `['*', '+my_attr']`

        """
        try:
            self._fields
        except AttributeError:
            fields = request.params.get('fields', None)
            if fields is not None:
                fields = json.loads(fields)
            self._fields = fields
        return self._fields

    def _get_wrap(self):
        try:
            self._wrap
        except AttributeError:
            self._set_wrap()
        return self._wrap
    def _set_wrap(self, value=None):
        """Set whether to wrap a template in its parent template.

        If ``value`` is given, use that value. If ``wrap`` is set in
        ``request.params``, convert its value to ``bool`` and use that
        value. Otherwise, set ``self.wrap`` to True.

        ``value``
            ``bool`` indicating whether or not a template should be
            wrapped in its inherited templates. ``wrap`` gets passed along
            to the template, which can decide to do whatever it wants with
            it.

        """
        if value is None:
            wrap = request.params.get('wrap', 'true')
            self._wrap = asbool(wrap)
        else:
            self._wrap = value
        c.wrap = self._wrap
    wrap = property(_get_wrap, _set_wrap)

    def __setattr__(self, name, value):
        """Set attribute on both ``self`` and ``c``."""
        attr = getattr(self.__class__, name, None)
        if isinstance(attr, property) and attr.fset is not None:
            # I.e., just call the property's _set method
            super(Controller, self).__setattr__(name, value)
        else:
            self._set_property([name], value)

    def _set_property(self, names, value):
        """Set attributes on both ``self`` and ``c``."""
        for name in names:
            self.__dict__[name] = value
            # Don't put "private" names in the template context
            if not name.startswith('_'):
                setattr(c, name, value)

    def _p_get_collection(self):
        return self.__dict__.get('collection', None)
    def _p_set_collection(self, collection):
        self._set_property(
            ['collection', self.collection_name], collection)
    collection = property(_p_get_collection, _p_set_collection)

    def _p_get_member(self):
        return self.__dict__.get('member', None)
    def _p_set_member(self, member):
        self._set_property(['member', self.member_name], member)
    member = property(_p_get_member, _p_set_member)

    def _init_properties(self):
        cls = self.__class__
        setattr(cls, self.collection_name, cls.collection)
        setattr(cls, self.member_name, cls.member)



from setuptools import setup, find_packages


version = '0.6.1'


setup(
    name='Restler',
    version=version,
    description="""RESTful base controller for Pylons 1.0.""",
    long_description="""\
Restler is a controller for Pylons projects that provides a set of default
RESTful actions that can be overridden as needed.

The Restler project is hosted at Google Code. Please see
http://code.google.com/p/restler/ for more details, documentation, etc.

Restler was originally extracted from the byCycle bicycle trip planner
(http://bycycle.org).

""",
    license='BSD/MIT',
    author='Wyatt L Baldwin',
    author_email='self@wyattbaldwin.com',
    keywords='web pylons controller REST WSGI',
    url='http://code.google.com/p/restler/',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Pylons',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=(
        'decorator>=3.1.2',
        'SQLAlchemy>=0.6.0',
    ),
    test_suite = 'nose.collector',
)




"""Entity base class.

The :class:`Entity` class here should be used as the base class for any and
all entity classes, regardless of what database the entities are derived from.

"""
import datetime
import decimal
try:
    import json
except ImportError:
    import simplejson as json

from string import ascii_uppercase

from sqlalchemy import Column
from sqlalchemy.orm import class_mapper


datetime_types = (datetime.time, datetime.date, datetime.datetime)


def underscore_to_title(name):
    return name.replace('_', ' ').title()


def camel_to_underscore(name):
    for char in ascii_uppercase:
        name = name.replace(char, '_%s' % char)
    name = name.lower()
    name = name.strip('_')
    return name


class Entity(object):

    @property
    def id(self):
        pk = self._sa_instance_state.key
        if pk is None:
            return None
        vals = pk[1]
        if len(vals) == 1:
            return vals[0]
        else:
            return tuple(vals)

    @property
    def id_str(self):
        """Convert `id` from Python to string."""
        if isinstance(self.id, basestring):
            id = self.id
        else:
            id = json.dumps(self.simplify_object(self.id))
        return id

    @classmethod
    def str_to_id(cls, id):
        """Convert ``id`` from string to Python.

        If `id` is scalar, return ``id`` as is; if it's a list, assume it's
        JSON encoded and decode it.

        """
        pk = class_mapper(cls).primary_key
        if cls.has_multipart_primary_key():
            id = json.loads(id)
            if not isinstance(id, list):
                raise ValueError(
                    'Expected a string that could be parsed as a multi-part '
                    'primary key--something like [1, "a"].')
            for i, col, val in zip(range(len(id)), pk, id):
                id[i] = cls.convert_param(col.name, val)
        else:
            id = cls.convert_param(pk[0].name, id)
        return id

    @classmethod
    def has_multipart_primary_key(cls):
        try:
            cls._has_multipart_primary_key
        except AttributeError:
            pk = class_mapper(cls).primary_key
            cls._has_multipart_primary_key = (True if (len(pk) > 1) else False)
        return cls._has_multipart_primary_key

    @classmethod
    def convert_param(self, name, val):
        return val

    @classmethod
    def simplify_object(cls, obj, name=None):
        """Convert ``obj`` to something JSON encoder can handle."""
        try:
            obj.to_simple_object
        except AttributeError:
            pass
        else:
            obj = obj.to_simple_object()
        if isinstance(obj, (list, tuple)):
            obj = [cls.simplify_object(i) for i in obj]
        elif isinstance(obj, decimal.Decimal):
            f, i = float(obj), int(obj)
            obj = i if f == i else f
        elif isinstance(obj, datetime_types):
            obj = str(obj)
        return obj

    def to_simple_object(self, fields=None):
        """Convert and return simplified form of `self` that can be JSONified.

        See :meth:`_parse_fields_for_simple_object` for description of
        ``fields``.

        """
        obj = dict(
            __module__=self.__class__.__module__,
            __type__=self.__class__.__name__,
        )
        include_fields = self._parse_fields_for_simple_object(fields)
        for name, as_name in include_fields:
            name_parts = name.split('.')
            o = self
            for n in name_parts:
                o = getattr(o, n)
            val = self.simplify_object(o, n)
            if name == as_name:
                # If `name` has only one part, this sets obj[name] = val.
                # If `name` has more than one part (N parts), this sets
                # obj[name1][name2][...][nameN] = val.
                slot = obj
                for n in name_parts[:-1]:
                    slot = slot.setdefault(n, {})
                slot[name_parts[-1]] = val
            else:
                # Use user-specified name
                obj[as_name] = val
        return obj

    def _parse_fields_for_simple_object(self, fields):
        """Parse ``fields`` and return the set of fields to be included.

        Each item in the returned set will be a 2-tuple mapping an Entity
        instance attribute name to another name. The mapped name might be the
        same as the attribute name::

            {('attr1', 'attr1'), ('attr2', 'different_name'), ...}

        ``fields`` is a list of attributes to include and/or exclude in the
        returned object object. If ``fields`` is `None`, the default set of
        fields will be used; (the list of names returned by
        :meth:`_public_names`).

        When ``fields`` is *not* `None`, the special value "*" can be given as
        one of the list items to indicate that the default set of fields
        should be used. In this case, "+" and "-" fields (see below) will be
        included or excluded relative to the default set.

        Each item in the list may be either a string or a dict. A string is
        used to indicate that an attribute's name should be used as-is in the
        response.

        A dict is used to map the field's column name to a different name
        in the response. In this case, the dict must contain a ``name`` key
        with the field's column name and a ``mapping`` key with the name
        desired in the reponse.

        In either case, the first character of the attribute name may be a "+"
        or a "-". These are used to, respectively, include or exclude fields
        from the result object. "-my_attr" used with "*" is an easy way to say
        "Everything but my_attr." "+my_attr" used with "*" is an easy way
        (the *only* way!) to include fields that aren't part of the default
        set (for example, relations are never included by default).

        """
        if fields is None:
            fields = ['*']
        elif isinstance(fields, dict):
            # Support legacy dict format wherein every field is mapped to
            # a name, even if it's the same name.
            fields = [dict(name=k, mapping=v) for k, v in fields.items()]
        mapped_fields = []
        include_fields = set()
        exclude_fields = set()
        for item in fields:
            if isinstance(item, basestring):
                name, as_name = item, item
            elif isinstance(item, dict):
                name, as_name = item['name'], item['mapping']
            if name.startswith('-'):
                exclude_fields.add(name.lstrip('-'))
            elif name == '*':
                mapped_fields += [(n, n) for n in self._public_names]
            else:
                # Note that unprefixed fields and fields prefixed with "+"
                # have the same semantics, and that's why we always strip
                # leading "+"s here.
                mapped_fields.append((name.lstrip('+'), as_name.lstrip('+')))
        for name, as_name in mapped_fields:
            if name not in exclude_fields:
                include_fields.add((name, as_name))
        return include_fields

    def to_json(self, fields=None):
        return json.dumps(self.to_simple_object(fields=fields))

    @classmethod
    def to_simple_collection(cls, collection, fields=None):
        try:
            collection[0].to_simple_object
        except AttributeError:
            # Assume collection of `RowTuple`s
            dicts = []
            for m in collection:
                keys = m.keys()
                vals = (getattr(m, k) for k in keys)
                dicts.append(dict(zip(keys, vals)))
            return [cls.simplify_object(d) for d in dicts]
        else:
            # Assume collection of instances of a mapped class
            return [m.to_simple_object(fields) for m in collection]

    @classmethod
    def to_json_collection(cls, collection=None, fields=None):
        simple_obj = cls.to_simple_collection(collection, fields=fields)
        return json.dumps(simple_obj)

    @property
    def _public_names(self):
        """We want all public DB columns and `property`s by default."""
        try:
            self.__public_names
        except AttributeError:
            names = []
            class_attrs = self.__class__.__dict__
            for name in class_attrs:
                if name.startswith('_'):
                    continue
                attr = class_attrs[name]
                if isinstance(attr, property):
                    names.append(name)
                else:
                    try:
                        clause_el = attr.__clause_element__()
                    except AttributeError:
                        pass
                    else:
                        if issubclass(clause_el.__class__, Column):
                            names.append(name)
            names = set(names)
            self.__public_names = names
        return self.__public_names

    def __str__(self):
        names = sorted(self._public_names)
        title = self.member_title
        string = [title, '-' * len(title)]
        string += ['%s: %s' % (name, getattr(self, name)) for name in names]
        return '\n'.join(string)


def instrument_class(cls, mixin=Entity):
    """Add member/collection class attributes to `Entity` subclass ``cls``.

    ``cls`` will be instrumented with ``member_name``, ``collection_name``,
    ``member_title``, and ``collection_title`` class attributes, IFF these
    class attributes are not already set in the class definition.

    In addition, ``mixin`` will be dynamically inserted into ``cls``'s
    list of base classes. ``mixin`` must be :class:`Entity` or a subclass
    (although this isn't currently enforced). ``mixin`` can also be `None`
    to completely disable this "feature" (XXX: Should this be the default?).

    """
    if not hasattr(cls, 'member_name'):
        cls.member_name = camel_to_underscore(cls.__name__)
    if not hasattr(cls, 'collection_name'):
        cls.collection_name = '{0}s'.format(cls.member_name)
    if not hasattr(cls, 'member_title'):
        cls.member_title = underscore_to_title(cls.member_name)
    if not hasattr(cls, 'collection_title'):
        cls.collection_title = underscore_to_title(cls.collection_name)
    if mixin is not None and mixin not in cls.__bases__:
            cls.__bases__ += (mixin,)
    return cls
